@@ -28,8 +28,26 @@
  *
  *  You should have received a copy of the GNU Lesser General Public
  *  License along with this program; if not, write to the Free
- *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  */
+ *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
+ *
+ * Modified BSD Licence
+ *
+ * Redistribution and use in source and binary forms, with or without
+ * modification, are permitted provided that the following conditions
+ * are met:
+ *
+ *   1. Redistributions of source code must retain the above copyright
+ *      notice, this list of conditions and the following disclaimer.
+ *   2. Redistributions in binary form must reproduce the above
+ *      copyright notice, this list of conditions and the following
+ *      disclaimer in the documentation and/or other materials provided
+ *      with the distribution.
+ *   3. The name of the author may not be used to endorse or promote
+ *      products derived from this software without specific prior
+ *      written permission.
+ */
 
+#include <config.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
@@ -41,6 +59,9 @@
 #include <sys/ioctl.h>
 #include <termios.h>
 #include <unistd.h>
+#include <sys/socket.h>
+#include <netinet/in.h>
+#include <netinet/tcp.h>
 
 #include <OpenIPMI/serv.h>
 #include <OpenIPMI/mcserv.h>
@@ -55,6 +76,26 @@
 #define SOL_INBUF_SIZE 32
 #define SOL_OUTBUF_SIZE 16384
 
+#define SOL_TELNET_IAC		255
+#define SOL_TELNET_DONT		254
+#define SOL_TELNET_DO		253
+#define SOL_TELNET_WONT		252
+#define SOL_TELNET_WILL		251
+#define SOL_TELNET_SB		250
+#define SOL_TELNET_SE		240
+#define SOL_TELNET_BREAK	243
+
+#define SOL_TELNETCODE_BINARY	0
+#define SOL_TELNETCODE_ECHO	1
+
+enum sol_telnet_state {
+    SOL_TELNETST_NORMAL = 0,
+    SOL_TELNETST_IAC,
+    SOL_TELNETST_TWOBYTE,
+    SOL_TELNETST_SUBNEG,
+    SOL_TELNETST_SUBNEG_IAC
+};
+
 struct soldata_s {
     int fd;
     sys_data_t *sys;
@@ -62,7 +103,8 @@ struct soldata_s {
     struct termios termctl;
     int modemstate;
 
-    channel_t *channel;
+    channel_t *channel; /* Channel for I/O */
+    channel_t *logchan; /* Channel for logging errors. */
     msg_t dummy_send_msg;
 
     /* Data from the remote to the serial port */
@@ -117,14 +159,517 @@ struct soldata_s {
     int curr_packet_seq;
     ipmi_timer_t *timer;
     unsigned int num_sends;
+
+    enum sol_telnet_state telnet_state;
+    unsigned char telnet_twobyte;
+
+    void (*reset_modemstate)(ipmi_sol_t *sol);
+    void (*update_bitrate)(ipmi_sol_t *sol);
+    void (*send_break)(ipmi_sol_t *sol);
+    void (*update_modemstate)(ipmi_sol_t *sol, int ctspaus, int deassertdcd);
+    int (*activate)(ipmi_sol_t *sol, msg_t *msg);
+    void (*shutdown)(ipmi_sol_t *sol);
+    int (*initialize)(ipmi_sol_t *sol);
 };
 
+#ifdef USE_UUCP_LOCKING
+static char *uucp_lck_dir = "/var/lock";
+static char *progname = "ipmisim";
+
+static int
+uucp_fname_lock_size(const char *devname)
+{
+    const char *ptr;
+
+    (ptr = strrchr(devname, '/'));
+    if (ptr == NULL) {
+	ptr = devname;
+    } else {
+	ptr = ptr + 1;
+    }
+
+    return 7 + strlen(uucp_lck_dir) + strlen(ptr);
+}
+
+static void
+uucp_fname_lock(char *buf, const char *devname)
+{
+    const char *ptr;
+
+    (ptr = strrchr(devname, '/'));
+    if (ptr == NULL) {
+	ptr = devname;
+    } else {
+	ptr = ptr + 1;
+    }
+    sprintf(buf, "%s/LCK..%s", uucp_lck_dir, ptr);
+}
+
+static int
+write_full(int fd, char *data, size_t count)
+{
+    ssize_t written;
+
+ restart:
+    while ((written = write(fd, data, count)) > 0) {
+	data += written;
+	count -= written;
+    }
+    if (written < 0) {
+	if (errno == EAGAIN)
+	    goto restart;
+	return -1;
+    }
+    return 0;
+}
+
+static void
+uucp_rm_lock(sys_data_t *sys, const char *devname)
+{
+    char *lck_file;
+
+    lck_file = sys->alloc(sys, uucp_fname_lock_size(devname));
+    if (lck_file == NULL) {
+	return;
+    }
+    uucp_fname_lock(lck_file, devname);
+    unlink(lck_file);
+    sys->free(sys, lck_file);
+}
+
+/* return 0=OK, -1=error, 1=locked by other proces */
+static int
+uucp_mk_lock(sys_data_t *sys, const char *devname)
+{
+    struct stat stt;
+    int pid = -1;
+
+    if (stat(uucp_lck_dir, &stt) == 0) { /* is lock file directory present? */
+	char *lck_file;
+	union {
+	    uint32_t ival;
+	    char     str[64];
+	} buf;
+	int fd;
+
+	lck_file = sys->alloc(sys, uucp_fname_lock_size(devname));
+	if (lck_file == NULL)
+	    return -1;
+
+	uucp_fname_lock(lck_file, devname);
+
+	pid = 0;
+	if ((fd = open(lck_file, O_RDONLY)) >= 0) {
+	    int n;
+
+	    n = read(fd, &buf, sizeof(buf));
+	    close(fd);
+	    if( n == 4 ) 		/* Kermit-style lockfile. */
+		pid = buf.ival;
+	    else if (n > 0) {		/* Ascii lockfile. */
+		buf.str[n] = 0;
+		sscanf(buf.str, "%d", &pid);
+	    }
+
+	    if (pid > 0 && kill((pid_t)pid, 0) < 0 && errno == ESRCH) {
+		/* death lockfile - remove it */
+		unlink(lck_file);
+		sleep(1);
+		pid = 0;
+	    }
+	}
+
+	if (pid == 0) {
+	    int mask;
+	    int rv;
+
+	    mask = umask(022);
+	    fd = open(lck_file, O_WRONLY | O_CREAT | O_EXCL, 0666);
+	    umask(mask);
+	    if (fd >= 0) {
+		snprintf(buf.str, sizeof(buf), "%10ld\t%s\n",
+			 (long)getpid(), progname );
+		rv = write_full(fd, buf.str, strlen(buf.str));
+		close(fd);
+		if (rv < 0) {
+		    pid = -errno;
+		    unlink(lck_file);
+		}
+	    } else {
+		pid = -errno;
+	    }
+	}
+
+	sys->free(sys, lck_file);
+    }
+
+    return pid;
+}
+#endif /* USE_UUCP_LOCKING */
+
+static int
+sol_to_termios_bitrate(ipmi_sol_t *sol, int solbps)
+{
+    int retried = 0;
+
+  retry:
+    switch(solbps) {
+    case 6: return B9600;
+    case 7: return B19200;
+    case 8: return B38400;
+    case 9: return B57600;
+    case 10: return B115200;
+
+    case 0:
+    default:
+	if (retried)
+	    return B9600;
+	solbps = sol->solparm.default_bitrate;
+	goto retry;
+    }
+}
+
+static void
+sol_serial_reset_modemstate(ipmi_sol_t *sol)
+{
+    int modemstate;
+
+    /* Turn on CTS and DCD if we have history, off if not */
+    /* Assuming standard NULL modem, RTS->CTS, DTR->DSR/DCD */
+    ioctl(sol->soldata->fd, TIOCMGET, &modemstate);
+    if (sol->history_size)
+	modemstate |= TIOCM_DTR | TIOCM_RTS;
+    else
+	modemstate &= ~(TIOCM_DTR | TIOCM_RTS);
+    sol->soldata->modemstate = modemstate & (TIOCM_DTR | TIOCM_RTS);
+    ioctl(sol->soldata->fd, TIOCMSET, &modemstate);
+}
+
+#ifndef HAVE_CFMAKERAW
+static void cfmakeraw(struct termios *t)
+{
+    t->c_iflag &= ~(IGNBRK | BRKINT | PARMRK | ISTRIP
+                    | INLCR | IGNCR | ICRNL | IXON);
+    t->c_oflag &= ~OPOST;
+    t->c_lflag &= ~(ECHO | ECHONL | ICANON | ISIG | IEXTEN);
+    t->c_cflag &= ~(CSIZE | PARENB);
+    t->c_cflag |= CS8;
+
+}
+#endif
+
+/* Initialize a serial port control structure for the first time. */
+static void
+devinit(ipmi_sol_t *sol, struct termios *termctl)
+{
+    int bitrate = sol_to_termios_bitrate(sol, sol->solparm.bitrate);
+
+    cfmakeraw(termctl);
+    cfsetospeed(termctl, bitrate);
+    cfsetispeed(termctl, bitrate);
+    termctl->c_cflag &= ~(CSTOPB);
+    termctl->c_cflag &= ~(CSIZE);
+    termctl->c_cflag |= CS8;
+    termctl->c_cflag &= ~(PARENB);
+    termctl->c_cflag |= CLOCAL;
+    termctl->c_cflag &= ~(HUPCL);
+    termctl->c_cflag |= CREAD;
+    if (sol->use_rtscts)
+	termctl->c_cflag |= CRTSCTS;
+    else
+	termctl->c_cflag &= ~(CRTSCTS);
+    termctl->c_iflag &= ~(IXON | IXOFF | IXANY);
+    termctl->c_iflag |= IGNBRK;
+
+    sol_serial_reset_modemstate(sol);
+}
+
+static void
+sol_serial_update_bitrate(ipmi_sol_t *sol)
+{
+    int bitrate = sol_to_termios_bitrate(sol, sol->solparm.bitrate);
+
+    cfsetospeed(&sol->soldata->termctl, bitrate);
+    cfsetispeed(&sol->soldata->termctl, bitrate);
+    tcsetattr(sol->soldata->fd, TCSANOW, &sol->soldata->termctl);
+}
+
+static void
+sol_serial_send_break(ipmi_sol_t *sol)
+{
+    soldata_t *sd = sol->soldata;
+
+    tcsendbreak(sd->fd, 0);
+}
+
+static void
+sol_serial_update_modemstate(ipmi_sol_t *sol, int ctspause, int deassert_dcd)
+{
+    soldata_t *sd = sol->soldata;
+
+    /*
+     * If history is enabled, we don't allow DCD/CTS fiddling, just leave
+     * them on all the time.
+     */
+    if (!sol->history_size) {
+	int modemstate = 0;
+	if (!ctspause)
+	    modemstate |= TIOCM_RTS;
+	if (!deassert_dcd)
+	    modemstate |= TIOCM_DTR;
+
+	if (modemstate != sd->modemstate) {
+	    int val;
+	    ioctl(sol->soldata->fd, TIOCMGET, &val);
+	    val &= ~(TIOCM_DTR | TIOCM_RTS);
+	    val |= modemstate;
+	    ioctl(sol->soldata->fd, TIOCMSET, &val);
+	}
+    }
+}
+
+static int
+sol_serial_activate(ipmi_sol_t *sol, msg_t *msg)
+{
+    soldata_t *sd = sol->soldata;
+
+    /*
+     * Note that we enable CTS and DCD if history is set, because we
+     * always monitor the history.
+     */
+    if (!sol->history_size) {
+	if ((msg->data[2] & 1) == 0) {
+	    /* Assuming standard NULL modem, RTS->CTS, DTR->DSR/DCD */
+	    int modemstate;
+	    ioctl(sd->fd, TIOCMGET, &modemstate);
+	    modemstate |= TIOCM_DTR | TIOCM_RTS;
+	    sd->modemstate = TIOCM_DTR | TIOCM_RTS;
+	    ioctl(sd->fd, TIOCMSET, &modemstate);
+	}
+    }
+
+    return 0;
+}
+
+static void
+sol_serial_shutdown(ipmi_sol_t *sol)
+{
+    soldata_t *sd = sol->soldata;
+
+    if (sd->fd >= 0)
+	close(sd->fd);
+    sd->fd = -1;
+
+#ifdef USE_UUCP_LOCKING
+    uucp_rm_lock(sd->sys, sol->device);
+#endif /* USE_UUCP_LOCKING */
+}
+
+static int
+sol_serial_initialize(ipmi_sol_t *sol)
+{
+    soldata_t *sd = sol->soldata;
+    int err;
+
+#ifdef USE_UUCP_LOCKING
+    err = uucp_mk_lock(sd->sys, sol->device);
+    if (err > 0) {
+	fprintf(stderr, "SOL device %s is already owned by process %d\n",
+		sol->device, err);
+	err = EBUSY;
+	goto out_nounlock;
+    }
+    if (err < 0) {
+	fprintf(stderr, "Error locking SOL device %s\n", sol->device);
+	err = -err;
+	goto out_nounlock;
+    }
+#endif /* USE_UUCP_LOCKING */
+    devinit(sol, &sd->termctl);
+
+    sd->fd = open(sol->device, O_NONBLOCK | O_NOCTTY | O_RDWR);
+    if (sd->fd == -1) {
+	err = errno;
+	fprintf(stderr, "Error opening SOL device %s\n", sol->device);
+	goto out;
+    }
+
+    err = tcsetattr(sd->fd, TCSANOW, &sd->termctl);
+    if (err == -1) {
+	err = errno;
+	close(sd->fd);
+	sd->fd = -1;
+	fprintf(stderr, "Error configuring SOL device %s\n", sol->device);
+	goto out;
+    }
+
+    /* Turn off BREAK. */
+    ioctl(sd->fd, TIOCCBRK);
+
+ out:
+#ifdef USE_UUCP_LOCKING
+    if (err)
+	uucp_rm_lock(sd->sys, sol->device);
+ out_nounlock:
+#endif
+    return err;
+}
+
+static void
+sol_serial_setup(ipmi_sol_t *sol)
+{
+    soldata_t *sd = sol->soldata;
+
+    sd->reset_modemstate = sol_serial_reset_modemstate;
+    sd->update_bitrate = sol_serial_update_bitrate;
+    sd->send_break = sol_serial_send_break;
+    sd->update_modemstate = sol_serial_update_modemstate;
+    sd->activate = sol_serial_activate;
+    sd->shutdown = sol_serial_shutdown;
+    sd->initialize = sol_serial_initialize;
+}
+
+static void
+sol_tcp_reset_modemstate(ipmi_sol_t *sol)
+{
+}
+
+static void
+sol_tcp_update_bitrate(ipmi_sol_t *sol)
+{
+}
+
+static void
+sol_tcp_send_break(ipmi_sol_t *sol)
+{
+    soldata_t *sd = sol->soldata;
+    int left = sizeof(sd->inbuf) - sd->inlen;
+
+     if (left < 2)
+	return;
+
+    sd->inbuf[sd->inlen++] = SOL_TELNET_IAC;
+    sd->inbuf[sd->inlen++] = SOL_TELNET_BREAK;
+}
+
+static void
+sol_tcp_update_modemstate(ipmi_sol_t *sol, int ctspause, int deassert_dcd)
+{
+}
+
+static int
+sol_tcp_activate(ipmi_sol_t *sol, msg_t *msg)
+{
+    return 0;
+}
+
+static void
+sol_tcp_shutdown(ipmi_sol_t *sol)
+{
+    soldata_t *sd = sol->soldata;
+
+    if (sd->fd >= 0)
+	close(sd->fd);
+    sd->fd = -1;
+}
+
+static int
+sol_tcp_initialize(ipmi_sol_t *sol)
+{
+    soldata_t *sd = sol->soldata;
+    struct addrinfo hints, *addr;
+    int rv;
+    int options;
+
+    if (sd->fd != -1) {
+	sd->logchan->log(sd->logchan, OS_ERROR, NULL,
+			 "Error sol activate on active port %s:%s",
+			 sol->tcpdest, sol->tcpport);
+	return -1;
+    }
+	
+    memset(&hints, 0, sizeof(hints));
+    hints.ai_family = AF_UNSPEC;
+    hints.ai_socktype = SOCK_STREAM;
+    rv = getaddrinfo(sol->tcpdest, sol->tcpport, &hints, &addr);
+    if (rv != 0) {
+	sd->logchan->log(sd->logchan, OS_ERROR, NULL,
+			 "Error getting tcp sol port address for %s:%s: %s",
+			 sol->tcpdest, sol->tcpport, gai_strerror(rv));
+	return -1;
+    }
+
+    sd->fd = socket(addr->ai_family, SOCK_STREAM, 0);
+    if (sd->fd == -1) {
+	rv = -1;
+	sd->logchan->log(sd->logchan, OS_ERROR, NULL,
+			 "Error creating tcp sol port socket for %s:%s: %s",
+			 sol->tcpdest, sol->tcpport, strerror(errno));
+	goto out;
+    }
+
+    rv = connect(sd->fd, addr->ai_addr, addr->ai_addrlen);
+    if (rv == -1) {
+	close(sd->fd);
+	sd->fd = -1;
+	if (sd->sys->debug & DEBUG_SOL)
+	    sd->logchan->log(sd->logchan, OS_ERROR, NULL,
+			   "Error connecting tcp sol port socket for %s:%s: %s",
+			   sol->tcpdest, sol->tcpport, strerror(errno));
+	goto out;
+    }
+
+    options = 1;
+    rv = setsockopt(sd->fd, IPPROTO_TCP, TCP_NODELAY,
+		    (char *) &options, sizeof(options));
+    if (rv == -1) {
+	close(sd->fd);
+	sd->fd = -1;
+	sd->logchan->log(sd->logchan, OS_ERROR, NULL,
+			 "Error setting nodelay on tcp sol port socket"
+			 " for %s:%s: %s",
+			 sol->tcpdest, sol->tcpport, strerror(errno));
+	goto out;
+    }
+
+    rv = fcntl(sd->fd, F_SETFL, O_NONBLOCK);
+    if (rv == -1) {
+	close(sd->fd);
+	sd->fd = -1;
+	sd->logchan->log(sd->logchan, OS_ERROR, NULL,
+			 "Error setting nonblock on tcp sol port socket"
+			 " for %s:%s: %s",
+			 sol->tcpdest, sol->tcpport, strerror(errno));
+	goto out;
+    }
+
+ out:	    
+    freeaddrinfo(addr);
+    return rv;
+}
+
+static void
+sol_tcp_setup(ipmi_sol_t *sol)
+{
+    soldata_t *sd = sol->soldata;
+
+    sd->reset_modemstate = sol_tcp_reset_modemstate;
+    sd->update_bitrate = sol_tcp_update_bitrate;
+    sd->send_break = sol_tcp_send_break;
+    sd->update_modemstate = sol_tcp_update_modemstate;
+    sd->activate = sol_tcp_activate;
+    sd->shutdown = sol_tcp_shutdown;
+    sd->initialize = sol_tcp_initialize;
+}
+
 static char *end_history_msg = "\r\n<End Of History>\r\n";
 
 #define MAX_HISTORY_SEND 64
 #define MAX_SOL_RESENDS 4
 static void sol_timeout(void *cb_data);
 static void sol_history_timeout(void *cb_data);
+static void sol_data_ready(int fd, void *cb_data);
+static void sol_write_ready(int fd, void *cb_data);
 
 static void sol_set_history_return_size(lmc_data_t    *mc,
 					msg_t         *msg,
@@ -178,22 +723,6 @@ static void sol_get_history_return_size(lmc_data_t    *mc,
     *rdata_len = 2;
 }
 
-static void
-reset_modem_state(ipmi_sol_t *sol)
-{
-    int modemstate;
-
-    /* Turn on CTS and DCD if we have history, off if not */
-    /* Assuming standard NULL modem, RTS->CTS, DTR->DSR/DCD */
-    ioctl(sol->soldata->fd, TIOCMGET, &modemstate);
-    if (sol->history_size)
-	modemstate |= TIOCM_DTR | TIOCM_RTS;
-    else
-	modemstate &= ~(TIOCM_DTR | TIOCM_RTS);
-    sol->soldata->modemstate = modemstate & (TIOCM_DTR | TIOCM_RTS);
-    ioctl(sol->soldata->fd, TIOCMSET, &modemstate);
-}
-
 static void
 sol_session_closed(lmc_data_t *mc, uint32_t session_id, void *cb_data)
 {
@@ -206,8 +735,9 @@ sol_session_closed(lmc_data_t *mc, uint32_t session_id, void *cb_data)
 	    sol->soldata->dummy_send_msg.src_addr = NULL;
 	}
 	sol->active = 0;
+	sd->channel = NULL;
 	sol->session_id = 0;
-	reset_modem_state(sol);
+	sd->reset_modemstate(sol);
     } else if (session_id == sol->history_session_id) {
 	if (sol->soldata->history_dummy_send_msg.src_addr) {
 	    sd->sys->free(sd->sys,
@@ -367,19 +897,13 @@ ipmi_sol_activate(lmc_data_t    *mc,
     dmsg->sid = msg->sid;
 
     if (instance == 1) {
-	/*
-	 * Note that we enable CTS and DCD if history is set, because we
-	 * always monitor the history.
-	 */
-	if (!sol->history_size) {
-	    if ((msg->data[2] & 1) == 0) {
-		/* Assuming standard NULL modem, RTS->CTS, DTR->DSR/DCD */
-		int modemstate;
-		ioctl(sd->fd, TIOCMGET, &modemstate);
-		modemstate |= TIOCM_DTR | TIOCM_RTS;
-		sd->modemstate = TIOCM_DTR | TIOCM_RTS;
-		ioctl(sd->fd, TIOCMSET, &modemstate);
-	    }
+	rv = sd->activate(sol, msg);
+	if (rv) {
+	    sd->sys->free(sd->sys, dmsg->src_addr);
+	    dmsg->src_addr = NULL;
+	    rdata[0] = IPMI_UNKNOWN_ERR_CC;
+	    *rdata_len = 1;
+	    return;
 	}
 
 	sol->active = 1;
@@ -456,198 +980,13 @@ ipmi_sol_deactivate(lmc_data_t    *mc,
     *rdata_len = 1;
 }
 
-#ifdef USE_UUCP_LOCKING
-static char *uucp_lck_dir = "/var/lock";
-static char *progname = "ipmisim";
-
-static int
-uucp_fname_lock_size(const char *devname)
-{
-    const char *ptr;
-
-    (ptr = strrchr(devname, '/'));
-    if (ptr == NULL) {
-	ptr = devname;
-    } else {
-	ptr = ptr + 1;
-    }
-
-    return 7 + strlen(uucp_lck_dir) + strlen(ptr);
-}
-
-static void
-uucp_fname_lock(char *buf, const char *devname)
-{
-    const char *ptr;
-
-    (ptr = strrchr(devname, '/'));
-    if (ptr == NULL) {
-	ptr = devname;
-    } else {
-	ptr = ptr + 1;
-    }
-    sprintf(buf, "%s/LCK..%s", uucp_lck_dir, ptr);
-}
-
-static int
-write_full(int fd, char *data, size_t count)
-{
-    size_t written;
-
- restart:
-    while ((written = write(fd, data, count)) > 0) {
-	data += written;
-	count -= written;
-    }
-    if (written < 0) {
-	if (errno == EAGAIN)
-	    goto restart;
-	return -1;
-    }
-    return 0;
-}
-
-static void
-uucp_rm_lock(sys_data_t *sys, const char *devname)
-{
-    char *lck_file;
-
-    lck_file = sys->alloc(sys, uucp_fname_lock_size(devname));
-    if (lck_file == NULL) {
-	return;
-    }
-    uucp_fname_lock(lck_file, devname);
-    unlink(lck_file);
-    sys->free(sys, lck_file);
-}
-
-/* return 0=OK, -1=error, 1=locked by other proces */
-static int
-uucp_mk_lock(sys_data_t *sys, const char *devname)
-{
-    struct stat stt;
-    int pid = -1;
-
-    if (stat(uucp_lck_dir, &stt) == 0) { /* is lock file directory present? */
-	char *lck_file;
-	union {
-	    uint32_t ival;
-	    char     str[64];
-	} buf;
-	int fd;
-
-	lck_file = sys->alloc(sys, uucp_fname_lock_size(devname));
-	if (lck_file == NULL)
-	    return -1;
-
-	uucp_fname_lock(lck_file, devname);
-
-	pid = 0;
-	if ((fd = open(lck_file, O_RDONLY)) >= 0) {
-	    int n;
-
-    	    n = read(fd, &buf, sizeof(buf));
-	    close(fd);
-	    if( n == 4 ) 		/* Kermit-style lockfile. */
-		pid = buf.ival;
-	    else if (n > 0) {		/* Ascii lockfile. */
-		buf.str[n] = 0;
-		sscanf(buf.str, "%d", &pid);
-	    }
-
-	    if (pid > 0 && kill((pid_t)pid, 0) < 0 && errno == ESRCH) {
-		/* death lockfile - remove it */
-		unlink(lck_file);
-		sleep(1);
-		pid = 0;
-	    }
-	}
-
-	if (pid == 0) {
-	    int mask;
-	    size_t rv;
-
-	    mask = umask(022);
-	    fd = open(lck_file, O_WRONLY | O_CREAT | O_EXCL, 0666);
-	    umask(mask);
-	    if (fd >= 0) {
-		snprintf(buf.str, sizeof(buf), "%10ld\t%s\n",
-			 (long)getpid(), progname );
-		rv = write_full(fd, buf.str, strlen(buf.str));
-		close(fd);
-		if (rv < 0) {
-		    pid = -errno;
-		    unlink(lck_file);
-		}
-	    } else {
-		pid = -errno;
-	    }
-	}
-
-	sys->free(sys, lck_file);
-    }
-
-    return pid;
-}
-#endif /* USE_UUCP_LOCKING */
-
-static int
-sol_to_termios_bitrate(ipmi_sol_t *sol, int solbps)
-{
-    int retried = 0;
-
-  retry:
-    switch(solbps) {
-    case 6: return B9600;
-    case 7: return B19200;
-    case 8: return B38400;
-    case 9: return B57600;
-    case 10: return B115200;
-
-    case 0:
-    default:
-	if (retried)
-	    return B9600;
-	solbps = sol->solparm.default_bitrate;
-	goto retry;
-    }
-}
-
-/* Initialize a serial port control structure for the first time. */
-static void
-devinit(ipmi_sol_t *sol, struct termios *termctl)
-{
-    int bitrate = sol_to_termios_bitrate(sol, sol->solparm.bitrate);
-
-    cfmakeraw(termctl);
-    cfsetospeed(termctl, bitrate);
-    cfsetispeed(termctl, bitrate);
-    termctl->c_cflag &= ~(CSTOPB);
-    termctl->c_cflag &= ~(CSIZE);
-    termctl->c_cflag |= CS8;
-    termctl->c_cflag &= ~(PARENB);
-    termctl->c_cflag |= CLOCAL;
-    termctl->c_cflag &= ~(HUPCL);
-    termctl->c_cflag |= CREAD;
-    if (sol->use_rtscts)
-	termctl->c_cflag |= CRTSCTS;
-    else
-	termctl->c_cflag &= ~(CRTSCTS);
-    termctl->c_iflag &= ~(IXON | IXOFF | IXANY);
-    termctl->c_iflag |= IGNBRK;
-
-    reset_modem_state(sol);
-}
-
 static void
 sol_update_bitrate(lmc_data_t *mc)
 {
     ipmi_sol_t *sol = ipmi_mc_get_sol(mc);
-    int bitrate = sol_to_termios_bitrate(sol, sol->solparm.bitrate);
+    soldata_t *sd = sol->soldata;
 
-    cfsetospeed(&sol->soldata->termctl, bitrate);
-    cfsetispeed(&sol->soldata->termctl, bitrate);
-    tcsetattr(sol->soldata->fd, TCSANOW, &sol->soldata->termctl);
+    sd->update_bitrate(ipmi_mc_get_sol(mc));
 }
 
 static void
@@ -734,6 +1073,50 @@ next_seq(soldata_t *sd)
 	sd->curr_packet_seq = 1;
 }
 
+static unsigned char sol_telnet_initseq[] = {
+    SOL_TELNET_IAC, SOL_TELNET_DO, SOL_TELNETCODE_BINARY,
+    SOL_TELNET_IAC, SOL_TELNET_WILL, SOL_TELNETCODE_BINARY,
+    SOL_TELNET_IAC, SOL_TELNET_DONT, SOL_TELNETCODE_ECHO,
+    SOL_TELNET_IAC, SOL_TELNET_WONT, SOL_TELNETCODE_ECHO
+};
+
+static int
+sol_port_init(ipmi_sol_t *sol)
+{
+    soldata_t *sd = sol->soldata;
+    int err;
+
+    err = sd->initialize(sol);
+    if (err)
+	goto out_err;
+
+    sol->configured++; /* Marked that we locked the device. */
+    sol->update_bitrate = sol_update_bitrate;
+
+    sd->read_enabled = 1;
+    sd->write_enabled = 0;
+
+    if (sol->do_telnet) {
+	int len = sizeof(sol_telnet_initseq);
+
+	memcpy(sd->inbuf, sol_telnet_initseq, len);
+	sd->inlen = len;
+    }
+
+    err = sd->sys->add_io_hnd(sd->sys, sd->fd, sol_data_ready, sol, &sd->fd_id);
+    if (err) {
+	sol->configured--;
+	sd->shutdown(sol);
+	goto out_err;
+    }
+
+    sd->sys->io_set_hnds(sd->fd_id, sol_write_ready, NULL);
+    set_write_enable(sd);
+
+ out_err:
+    return err;
+}
+
 static void
 sol_timeout(void *cb_data)
 {
@@ -741,16 +1124,25 @@ sol_timeout(void *cb_data)
     soldata_t *sd = sol->soldata;
     struct timeval tv;
 
-    if (sd->num_sends > MAX_SOL_RESENDS) {
-	sd->waiting_ack = 0;
-	next_seq(sd);
-	sd->outlen = 0;
-	return;
+    if (sd->fd == -1) {
+	sol_port_init(sol);
+    } else if (sol->active) {
+	if (sd->num_sends > MAX_SOL_RESENDS) {
+	    sd->waiting_ack = 0;
+	    next_seq(sd);
+	    sd->outlen = 0;
+	    return;
+	}
+
+	sd->num_sends++;
+	send_data(sol, 0);
     }
 
-    sd->num_sends++;
-    send_data(sol, 0);
-    tv.tv_sec = 1;
+    if (sd->fd == -1)
+	/* Attempt to reconnect every 10 seconds. */
+	tv.tv_sec = 10;
+    else
+	tv.tv_sec = 1;
     tv.tv_usec = 0;
     sd->sys->start_timer(sd->timer, &tv);
 }
@@ -786,14 +1178,37 @@ handle_sol_port_payload(lanserv_data_t *lan, ipmi_sol_t *sol, msg_t *msg)
     if (seq != 0) {
 	if (seq == sd->last_acked_packet) {
 	    need_send_ack = 1;
+	} else if (sd->fd == -1) {
+	    /* Ignore the data. */
+	    if  (len) {
+		sd->last_acked_packet = seq;
+		need_send_ack = 1;
+		sd->last_acked_packet_len = len;
+	    }
 	} else if (len) {
 	    sd->last_acked_packet = seq;
-	    if (len > (sizeof(sd->inbuf) - sd->inlen))
-		len = sizeof(sd->inbuf) - sd->inlen;
-	    sd->last_acked_packet_len = len;
-	    memcpy(sd->inbuf + sd->inlen, data, len);
-	    sd->inlen += len;
+	    if (sol->do_telnet) {
+		unsigned int i;
 
+		for (i = 0; i < len; i++) {
+		    int left = sizeof(sd->inbuf) - sd->inlen;
+		    if (left < 1)
+			break;
+		    if (data[i] == SOL_TELNET_IAC) {
+			if (left < 2)
+			    break;
+			sd->inbuf[sd->inlen++] = SOL_TELNET_IAC;
+		    }
+		    sd->inbuf[sd->inlen++] = data[i];
+		}
+	    } else {
+		if (len > (sizeof(sd->inbuf) - sd->inlen))
+		    len = sizeof(sd->inbuf) - sd->inlen;
+		memcpy(sd->inbuf + sd->inlen, data, len);
+		sd->inlen += len;
+	    }
+
+	    sd->last_acked_packet_len = len;
 	    need_send_ack = 1;
 	    set_write_enable(sol->soldata);
 	}
@@ -840,27 +1255,9 @@ handle_sol_port_payload(lanserv_data_t *lan, ipmi_sol_t *sol, msg_t *msg)
 	sd->inlen = 0;
 
     if (isbreak)
-	tcsendbreak(sd->fd, 0);
+	sd->send_break(sol);
 
-    /*
-     * If history is enabled, we don't allow DCD/CTS fiddling, just leave
-     * them on all the time.
-     */
-    if (!sol->history_size) {
-	int modemstate = 0;
-	if (!ctspause)
-	    modemstate |= TIOCM_RTS;
-	if (!deassert_dcd)
-	    modemstate |= TIOCM_DTR;
-
-	if (modemstate != sd->modemstate) {
-	    int val;
-	    ioctl(sol->soldata->fd, TIOCMGET, &val);
-	    val &= ~(TIOCM_DTR | TIOCM_RTS);
-	    val |= modemstate;
-	    ioctl(sol->soldata->fd, TIOCMSET, &val);
-	}
-    }
+    sd->update_modemstate(sol, ctspause, deassert_dcd);
 }
 
 static int
@@ -1007,6 +1404,24 @@ handle_sol_payload(lanserv_data_t *lan, msg_t *msg)
 	handle_sol_history_payload(lan, sol, msg);
 }
 
+static void
+sol_port_error(ipmi_sol_t *sol)
+{
+    soldata_t *sd = sol->soldata;
+    struct timeval tv;
+
+    sd->sys->remove_io_hnd(sd->fd_id);
+    sd->shutdown(sol);
+
+    /* Clear any data to send to the serial port. */
+    sd->inlen = 0;
+
+    /* Retry in 10 seconds. */
+    tv.tv_sec = 10;
+    tv.tv_usec = 0;
+    sd->sys->start_timer(sd->timer, &tv);
+}
+
 static void
 sol_write_ready(int fd, void *cb_data)
 {
@@ -1016,12 +1431,10 @@ sol_write_ready(int fd, void *cb_data)
 
     rv = write(fd, sd->inbuf, sd->inlen);
     if (rv < 0) {
-	sd->channel->log(sd->channel, OS_ERROR, NULL,
-			 "Error reading from serial port: %d, disabling\n",
+	sd->logchan->log(sd->logchan, OS_ERROR, NULL,
+			 "Error writing to serial port: %d, disabling\n",
 			 errno);
-	sd->sys->remove_io_hnd(sd->fd_id);
-	close(sd->fd);
-	sd->fd = -1;
+	sol_port_error(sol);
 	return;
     }
 
@@ -1083,6 +1496,65 @@ add_to_history(ipmi_sol_t *sol, unsigned char *buf, unsigned int len)
     sd->history_end += len;
 }
 
+static int
+sol_handle_telnet(ipmi_sol_t *sol, unsigned char *buf, int len)
+{
+    soldata_t *sd = sol->soldata;
+    int i, j;
+
+    for (i = 0, j = 0; i < len; i++) {
+	switch (sd->telnet_state) {
+	case SOL_TELNETST_NORMAL:
+	    if (buf[i] == SOL_TELNET_IAC)
+		sd->telnet_state = SOL_TELNETST_IAC;
+	    else
+		buf[j++] = buf[i];
+	    break;
+
+	case SOL_TELNETST_IAC:
+	    switch (buf[i]) {
+	    case SOL_TELNET_DONT:
+	    case SOL_TELNET_DO:
+	    case SOL_TELNET_WONT:
+	    case SOL_TELNET_WILL:
+		sd->telnet_twobyte = buf[i];
+		sd->telnet_state = SOL_TELNETST_TWOBYTE;
+		break;
+
+	    case SOL_TELNET_SB:
+		sd->telnet_state = SOL_TELNETST_SUBNEG;
+		break;
+
+	    case SOL_TELNET_IAC:
+		buf[j++] = SOL_TELNET_IAC;
+		/* Fallthrough */
+	    default:
+		/* Ignore everything else. */
+		sd->telnet_state = SOL_TELNETST_NORMAL;
+		break;
+	    }
+	    break;
+
+	case SOL_TELNETST_TWOBYTE:
+	    /* Don't listen to these. */
+	    sd->telnet_state = SOL_TELNETST_NORMAL;
+	    break;
+
+	case SOL_TELNETST_SUBNEG:
+	    /* Don't listen to these. */
+	    if (buf[i] == SOL_TELNET_IAC)
+		sd->telnet_state = SOL_TELNETST_SUBNEG_IAC;
+	    break;
+
+	case SOL_TELNETST_SUBNEG_IAC:
+	    if (buf[i] == SOL_TELNET_SE)
+		sd->telnet_state = SOL_TELNETST_NORMAL;
+	}
+    }
+
+    return j;
+}
+
 static void
 sol_data_ready(int fd, void *cb_data)
 {
@@ -1098,15 +1570,20 @@ sol_data_ready(int fd, void *cb_data)
 
     rv = read(fd, buf, readsize);
     if (rv < 0) {
-	sd->channel->log(sd->channel, OS_ERROR, NULL,
+	sd->logchan->log(sd->logchan, OS_ERROR, NULL,
 			 "Error reading from serial port: %d, disabling\n",
 			 errno);
-	sd->sys->remove_io_hnd(sd->fd_id);
-	close(sd->fd);
-	sd->fd = -1;
+	sol_port_error(sol);
+	return;
+    } else if (rv == 0) {
+	/* End of input, socket probably closed. */
+	sol_port_error(sol);
 	return;
     }
 
+    if (sol->do_telnet)
+	rv = sol_handle_telnet(sol, buf, rv);
+
     add_to_history(sol, buf, rv);
 
     if (((unsigned int) rv) > (sizeof(sd->outbuf) - sd->outlen)) {
@@ -1136,16 +1613,36 @@ sol_data_ready(int fd, void *cb_data)
 int
 sol_read_config(char **tokptr, sys_data_t *sys, const char **err)
 {
+    ipmi_sol_t *sol = sys->sol;
     unsigned int val;
     int          rv;
     const char   *tok;
 
     sys->sol->use_rtscts = 1;
 
-    rv = get_delim_str(tokptr, &sys->sol->device, err);
+    rv = get_delim_str(tokptr, &sol->device, err);
     if (rv)
 	return rv;
 
+    if (strncmp(sol->device, "tcp:", 4) == 0) {
+	sol->tcpdest = sol->device + 4;
+    } else if (strncmp(sol->device, "telnet:", 7) == 0) {
+	sol->do_telnet = 1;
+	sol->tcpdest = sol->device + 7;
+    }
+
+    if (sol->tcpdest) {
+	char *colon = strchr(sol->tcpdest, ':');
+
+	if (!colon) {
+	    *err = "No port specified in sol tcp connection";
+	    return -1;
+	}
+
+	sol->tcpport = colon + 1;
+	*colon = '\0';
+    }
+    
     rv = get_uint(tokptr, &val, err);
     if (rv)
 	return rv;
@@ -1163,7 +1660,7 @@ sol_read_config(char **tokptr, sys_data_t *sys, const char **err)
     while ((tok = mystrtok(NULL, " \t\n", tokptr))) {
 	if (strncmp(tok, "history=", 8) == 0) {
 	    char *end, next;
-	    sys->sol->history_size = strtoul(tok + 8, &end, 0);
+	    sol->history_size = strtoul(tok + 8, &end, 0);
 	    next = *end;
 	    while (next == ',') {
 		char *opt = end + 1;
@@ -1175,7 +1672,7 @@ sol_read_config(char **tokptr, sys_data_t *sys, const char **err)
 		*end = '\0';
 
 		if (strncmp(opt, "backupfile=", 11) == 0) {
-		    sys->sol->backupfile = strdup(opt + 11);
+		    sol->backupfile = strdup(opt + 11);
 		} else {
 		    *err = "Unknown history option";
 		    return -1;
@@ -1212,17 +1709,17 @@ sol_read_config(char **tokptr, sys_data_t *sys, const char **err)
 		return -1;
 	    }
 	} else if (strncmp(tok, "nortscts", 8) == 0) {
-	    sys->sol->use_rtscts = 0;
+	    sol->use_rtscts = 0;
 	} else if (strncmp(tok, "readclear", 8) == 0) {
-	    sys->sol->readclear = 1;
+	    sol->readclear = 1;
 	} else {
 	    *err = "Invalid item";
 	    return -1;
 	}
     }
 
-    sys->sol->solparm.default_bitrate = val;
-    sys->sol->configured = 1;
+    sol->solparm.default_bitrate = val;
+    sol->configured = 1;
     return 0;
 }
 
@@ -1322,9 +1819,8 @@ handle_sol_shutdown(void *info, int sig)
     if (sol->configured < 2 || !sd)
 	return;
 
-#ifdef USE_UUCP_LOCKING
-    uucp_rm_lock(sd->sys, sol->device);
-#endif /* USE_UUCP_LOCKING */
+    sol->configured--;
+    sd->shutdown(sol);
 
     if (!sol->backupfile || (sd->history_start == sd->history_end))
 	return;
@@ -1353,13 +1849,19 @@ sol_init_mc(sys_data_t *sys, lmc_data_t *mc)
 {
     ipmi_sol_t *sol = ipmi_mc_get_sol(mc);
     soldata_t *sd;
-    int err;
 
     sd = sys->alloc(sys, sizeof(*sd));
     if (!sd)
 	return ENOMEM;
     memset(sd, 0, sizeof(*sd));
+    sd->sys = sys;
+    sol->soldata = sd;
 
+    if (sol->tcpdest)
+	sol_tcp_setup(sol);
+    else
+	sol_serial_setup(sol);
+    
     if (sys->alloc_timer(sys, sol_timeout, sol, &sd->timer)) {
 	sys->free(sys, sd);
 	return ENOMEM;
@@ -1397,71 +1899,19 @@ sol_init_mc(sys_data_t *sys, lmc_data_t *mc)
     sd->backupfilehandler.info = sol;
     ipmi_register_shutdown_handler(&sd->backupfilehandler);
 
-    sd->sys = sys;
-    sol->soldata = sd;
     sd->fd = -1;
     sd->curr_packet_seq = 1;
     sd->history_curr_packet_seq = 1;
+    sd->logchan = ipmi_mc_get_channelset(mc)[0];
 
-#ifdef USE_UUCP_LOCKING
-    err = uucp_mk_lock(sys, sol->device);
-    if (err > 0) {
-	fprintf(stderr, "SOL device %s is already owned by process %d\n",
-		sol->device, err);
-	err = EBUSY;
-	goto out;
-    }
-    if (err < 0) {
-	fprintf(stderr, "Error locking SOL device %s\n", sol->device);
-	err = -err;
-	goto out;
-    }
-#endif /* USE_UUCP_LOCKING */
-    sol->configured++; /* Marked that we locked the device. */
+    if (sol_port_init(sol)) {
+	/* Retry in 10 seconds. */
+	struct timeval tv;
 
-    devinit(sol, &sd->termctl);
-
-    sd->fd = open(sol->device, O_NONBLOCK | O_NOCTTY | O_RDWR);
-    if (sd->fd == -1) {
-	err = errno;
-	fprintf(stderr, "Error opening SOL device %s\n", sol->device);
-	goto out;
+	tv.tv_sec = 10;
+	tv.tv_usec = 0;
+	sd->sys->start_timer(sd->timer, &tv);
     }
 
-    err = tcsetattr(sd->fd, TCSANOW, &sd->termctl);
-    if (err == -1) {
-	err = errno;
-	fprintf(stderr, "Error configuring SOL device %s\n", sol->device);
-	goto out;
-    }
-   
-    sol->update_bitrate = sol_update_bitrate;
-
-    /* Turn off BREAK. */
-    ioctl(sd->fd, TIOCCBRK);
-
-    sd->read_enabled = 1;
-    sd->write_enabled = 0;
-    err = sys->add_io_hnd(sys, sd->fd, sol_data_ready, sol, &sd->fd_id);
-    if (err)
-	goto out;
-
-    sd->sys->io_set_hnds(sd->fd_id, sol_write_ready, NULL);
-    set_write_enable(sd);
-
-  out:
-    if (err) {
-	if (sd->timer)
-	    sys->free_timer(sd->timer);
-	if (sd->history_timer)
-	    sys->free_timer(sd->history_timer);
-	if (sd->fd >= 0)
-	    close(sd->fd);
-	if (sd->history)
-	    sys->free(sys, sd->history);
-	sys->free(sys, sd);
-	sol->soldata = NULL;
-    }
-
-    return err;
+    return 0;
 }
